#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
 A fast and light-weight web vulnerability scanner. It helps pen-testers pinpoint possibly vulnerable targets from a large number of web servers.
 https://github.com/lijiejie/BBScan
 Li JieJie  my[at]lijiejie.com  https://www.lijiejie.com
"""

import os
# first, change working dir
cur_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(cur_dir)

import sys
import codecs
import asyncio
import httpx
import re
from bs4 import BeautifulSoup
import warnings
import time
import glob
import ipaddress
import ssl
import traceback
import importlib
import copy
import string
import random
import dns.asyncresolver
from urllib.parse import urlparse

from lib.common import clear_queue, parse_url, cal_depth, get_domain_sub, is_port_open, scan_given_ports, \
    is_ip_addr, get_dns_resolver, get_http_title, clear_url
from lib.cmdline import parse_args
from lib.report import save_report
import lib.config as conf
from lib.cms_fingerprints import Fingerprint
import hashlib
from lib.javascript_parser import get_urls_in_js_async


if hasattr(ssl, '_create_unverified_context'):
    ssl._create_default_https_context = ssl._create_unverified_context

from bs4 import MarkupResemblesLocatorWarning
warnings.filterwarnings('ignore', category=MarkupResemblesLocatorWarning)


fingerprint = Fingerprint()


class Scanner(object):
    def __init__(self, timeout=900):
        self.q_results = q_results
        self.args = args
        self.start_time = time.time()
        self.time_out = timeout
        self.links_limit = 100  # max number of folders allowed to scan

    async def init(self):
        await self._init_rules()
        self._init_scripts()

        self.url_queue = asyncio.Queue()  # all urls to scan
        self.urls_processed = set()     # processed urls
        self.urls_enqueued = set()      # entered queue urls
        self.urls_crawled = set()

        self.lock = asyncio.Lock()
        self.results = {}
        self.log_file = None
        self._404_status = -1
        self.conn_pool = None
        self.index_status, self.index_headers, self.index_html_doc = None, {}, ''
        self.scheme, self.host, self.port, self.path = None, None, None, None
        self.domain_sub = ''
        self.base_url = ''
        self.max_depth = 0
        self.len_404_doc = 0
        self.has_http = None
        self.ports_open = None
        self.ports_closed = None
        self.no_scripts = None
        self.status_502_count = 0
        self.timeout_count = 0
        self.timeout_scan_aborted = False
        self.fingerprint_check = True
        self.js_urls = []
        self.index_has_reported = False
        self.urls_regex_found = set()

    async def print_msg(self, msg):
        await self.q_results.put(msg)

    def reset_scanner(self):
        self.start_time = time.time()
        clear_queue(self.url_queue)
        self.urls_processed.clear()
        self.urls_enqueued.clear()
        self.urls_crawled.clear()
        self.results.clear()
        self.log_file = None
        self._404_status = -1
        # self.conn_pool = None    # Bug Fixed, shouldn't set to None right here, used pool can not be closed
        self.index_status, self.index_headers, self.index_html_doc = None, {}, ''
        self.scheme, self.host, self.port, self.path = None, None, None, None
        self.domain_sub = ''
        self.base_url = ''
        self.status_502_count = 0
        self.timeout_count = 0
        self.timeout_scan_aborted = False
        self.fingerprint_check = True
        self.js_urls = []
        self.index_has_reported = False
        self.urls_regex_found = set()

    # scan from a given URL
    async def init_from_url(self, target):
        self.reset_scanner()
        self.scheme = target['scheme']
        self.host = target['host']
        self.port = target['port']
        self.path = target['path']
        self.has_http = target['has_http']
        self.ports_open = target['ports_open']
        self.ports_closed = target['ports_closed']
        self.no_scripts = target['no_scripts'] if 'no_scripts' in target else 0
        self.domain_sub = get_domain_sub(self.host)
        await self.init_final()
        return True

    # Fix me: not yet implemented and tested 2024-05-27
    async def init_from_log_file(self, log_file):
        self.reset_scanner()
        self.log_file = log_file
        self.scheme, self.host, self.path = self._parse_url_from_file()
        self.domain_sub = get_domain_sub(self.host)
        if self.host:
            if self.host.find(':') > 0:
                _ret = self.host.split(':')
                self.host = _ret[0]
                self.port = _ret[1]
            elif self.scheme == 'https':
                self.port = 443
            elif self.scheme == 'http':
                self.port = 80
            else:
                self.port = None
            if await is_port_open(self.host, self.port):
                await self.print_msg('[Port Not Open] %s:%s' % (self.host, self.port))
                return False
            self.has_http = True
            self.no_scripts = 1
            await self.init_final()
            await self.load_all_urls_from_log_file()
            return True
        else:
            host = os.path.basename(log_file).replace('.log', '')
            try:
                await dns.asyncresolver.resolve(host, "A")
                await self.init_from_url(host)     # Fix Me
                return True
            except Exception as e:
                await self.print_msg('[ERROR] Invalid host from log name: %s' % host)
                return False

    async def init_final(self):
        try:
            if self.conn_pool:
                await self.conn_pool.aclose()
        except Exception as e:
            await self.print_msg('conn_pool.aclose exception: %s' % str(e))
        self.conn_pool = None    # after close
        if self.scheme == 'http' and self.port == 80 or self.scheme == 'https' and self.port == 443:
            self.base_url = '%s://%s' % (self.scheme, self.host)
        else:
            self.base_url = '%s://%s:%s' % (self.scheme, self.host, self.port)

        if self.has_http:
            await self.print_msg('Scan %s' % self.base_url)
        else:
            await self.print_msg('Scan %s:%s' % (self.host, self.port) if self.port else 'Scan %s' % self.host)

        if self.has_http:
            limits = httpx.Limits(max_connections=100, max_keepalive_connections=40)
            self.conn_pool = httpx.AsyncClient(headers=conf.default_headers,
                                               proxies=args.proxy, verify=False, limits=limits, follow_redirects=False)

            if self.args.require_index_doc:
                await self.crawl('/', do_not_process_links=True)

        if self.no_scripts != 1:   # 不是重复目标 80 443 跳转的，不需要重复扫描
            # 当前目标disable， 或者 全局开启插件扫描
            if self.args.scripts_only or not self.no_scripts:
                for _ in self.user_scripts:
                    await self.url_queue.put((_, '/'))

        if not self.has_http or self.args.scripts_only:    # 未发现HTTP服务 或  只依赖插件扫描
            return

        self.max_depth = cal_depth(self, self.path)[1] + 5
        if self.args.no_check404:
            self._404_status = 404
        else:
            await self.check_404_existence()
        if self._404_status == -1:
            await self.print_msg('[Warning] HTTP 404 check failed: %s' % self.base_url)
        # elif self._404_status != 404:
        #     await self.print_msg('[Warning] %s has no HTTP 404.' % self.base_url)
        _path, _depth = cal_depth(self, self.path)

        await self.enqueue('/')
        if _path != '/' and not self.log_file:
            await self.enqueue(_path)

    def _parse_url_from_file(self):
        url = ''
        with open(self.log_file) as infile:
            for _line in infile.readlines():
                _line = _line.strip()
                if _line and len(_line.split()) >= 3:
                    url = _line.split()[1]
                    break
        return parse_url(url)

    # load urls from rules/*.txt
    async def _init_rules(self):
        self.text_to_find = []
        self.regex_to_find = []
        self.text_to_exclude = []
        self.regex_to_exclude = []
        self.rules_set = set()
        self.rules_set_root_only = set()

        p_tag = re.compile('{tag="(.*?)"}')
        p_status = re.compile(r'{status=(\d{3})}')
        p_content_type = re.compile('{type="(.*?)"}')
        p_content_type_no = re.compile('{type_no="(.*?)"}')

        _files = self.args.rule_files if self.args.rule_files else glob.glob('rules/*.txt')
        if self.args.fingerprint_only:
            _files = []

        for rule_file in _files:
            with codecs.open(rule_file, 'r', encoding='utf-8') as infile:
                vul_type = os.path.basename(rule_file)[:-4]
                for url in infile.readlines():
                    url = url.strip()
                    if url.startswith('/'):
                        _ = p_tag.search(url)
                        tag = _.group(1) if _ else ''

                        _ = p_status.search(url)
                        status = int(_.group(1)) if _ else 0

                        _ = p_content_type.search(url)
                        content_type = _.group(1) if _ else ''

                        _ = p_content_type_no.search(url)
                        content_type_no = _.group(1) if _ else ''

                        root_only = True if url.find('{root_only}') >= 0 else False

                        rule = (url.split()[0], tag, status, content_type, content_type_no, root_only, vul_type)
                        if root_only:
                            if rule not in self.rules_set_root_only:
                                self.rules_set_root_only.add(rule)
                            else:
                                await self.print_msg('Duplicated root only rule: %s' % str(rule))
                        else:
                            if rule not in self.rules_set:
                                self.rules_set.add(rule)
                            else:
                                await self.print_msg('Duplicated rule: %s' % str(rule))

        re_text = re.compile('{text="(.*)"}')
        re_regex_text = re.compile('{regex_text="(.*)"}')

        file_path = 'rules/white.list'
        if not os.path.exists(file_path):
            await self.print_msg('[ERROR] File not exist: %s' % file_path)
            return
        for _line in codecs.open(file_path, encoding='utf-8'):
            _line = _line.strip()
            if not _line or _line.startswith('#'):
                continue
            _m = re_text.search(_line)
            if _m:
                self.text_to_find.append(_m.group(1))
            else:
                _m = re_regex_text.search(_line)
                if _m:
                    self.regex_to_find.append(re.compile(_m.group(1)))

        file_path = 'rules/black.list'
        if not os.path.exists(file_path):
            await self.print_msg('[ERROR] File not exist: %s' % file_path)
            return
        for _line in codecs.open(file_path, encoding='utf-8'):
            _line = _line.strip()
            if not _line or _line.startswith('#'):
                continue
            _m = re_text.search(_line)
            if _m:
                self.text_to_exclude.append(_m.group(1))
            else:
                _m = re_regex_text.search(_line)
                if _m:
                    self.regex_to_exclude.append(re.compile(_m.group(1)))

    def _init_scripts(self):
        self.user_scripts = []
        if self.args.no_scripts:    # 全局禁用插件，无需导入
            return
        files = 'scripts/*.py'
        if self.args.fingerprint_only:
            files = 'scripts/is_admin_site.py'
        for _script in glob.glob(files):
            script_name_origin = os.path.basename(_script)
            script_name = script_name_origin.replace('.py', '')
            if self.args.script:    # 只导入指定的脚本
                if script_name not in self.args.script and script_name_origin not in self.args.script:
                    continue
            if script_name.startswith('_'):
                continue
            try:
                self.user_scripts.append(importlib.import_module('scripts.%s' % script_name))
            except Exception as e:
                print('[ERROR] Fail to load script %s' % script_name)

    async def http_request(self, url, headers=conf.default_headers, timeout=30, follow_redirects=False):
        try:
            if not url:
                url = '/'
            if not self.conn_pool or self.timeout_scan_aborted:
                return -1, {}, ''
            if self.args.debug:
                await self.print_msg('--> %s' % self.base_url + url)
            resp = await self.conn_pool.get(self.base_url + url,
                                            headers=headers, follow_redirects=follow_redirects, timeout=timeout)
            if resp.headers.get('content-type', '').find('text') >= 0 \
                    or resp.headers.get('content-type', '').find('html') >= 0 \
                    or int(resp.headers.get('content-length', '0')) <= 20480:  # 1024 * 20
                html_doc = resp.text
            else:
                html_doc = ''

            if resp.status_code == 502:    # 502出现超过3次，排除该站点不再扫描
                self.status_502_count += 1
                if self.status_502_count > 3:
                    self.timeout_scan_aborted = True
                    clear_queue(self.url_queue)
                    try:
                        if self.conn_pool:
                            await self.conn_pool.aclose()
                    except Exception as e:
                        pass    #
                    self.conn_pool = None
                    if self.args.debug:
                        await self.print_msg('Website 502 exceeded: %s' % self.base_url)

            return resp.status_code, resp.headers, html_doc
        except httpx.ReadTimeout as e:
            self.timeout_count += 1
            if self.timeout_count >= 3:
                if not self.timeout_scan_aborted:
                    self.timeout_scan_aborted = True
                    await self.print_msg('[Warning] timeout exceeded, scan aborted: %s' % self.base_url)
                clear_queue(self.url_queue)
            return -1, {}, ''
        except (httpx.RequestError, httpx.HTTPStatusError, ssl.SSLError) as e:
            if self.args.debug:
                await self.print_msg('[Request Error] %s %s %s' % (type(e), str(e), self.base_url))
            return -1, {}, ''
        except Exception as e:
            if self.args.debug:
                await self.print_msg('[Request Error] %s %s %s' % (type(e), str(e), self.base_url))
            return -1, {}, ''

    async def check_404_existence(self):
        try:
            try:
                path = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in
                               range(random.randint(10, 30)))
                self._404_status, _, html_doc = await self.http_request('/' + path)
            except Exception as e:
                await self.print_msg('[Warning] HTTP 404 check failed: %s, %s' % (self.base_url, type(e)))
                self._404_status, _, html_doc = -1, {}, ''
            if self._404_status != 404:
                self.len_404_doc = len(html_doc)
        except Exception as e:
            await self.print_msg('[Check_404] Exception %s %s' % (self.base_url, str(e)))

    #
    async def enqueue(self, url):
        try:
            url = str(url)
        except Exception as e:
            return False
        try:
            url_pattern = re.sub(r'\d+', '{num}', url)
            if url_pattern in self.urls_processed or len(self.urls_processed) >= self.links_limit:
                return False

            self.urls_processed.add(url_pattern)
            # await self.print_msg('Entered Queue: %s' % url)
            if not self.args.no_crawl:   # no crawl
                await self.crawl(url)
            if self._404_status != -1:    # valid web service
                rule_set_to_process = [self.rules_set, self.rules_set_root_only] if url == '/' else [self.rules_set]
                for rule_set in rule_set_to_process:
                    for _ in rule_set:
                        if _[5] and url != '/':    # root only
                            continue
                        try:
                            full_url = url.rstrip('/') + _[0]
                        except Exception as e:
                            continue
                        if full_url in self.urls_enqueued:
                            continue
                        url_description = {'prefix': url.rstrip('/'), 'full_url': full_url}
                        item = (url_description, _[1], _[2], _[3], _[4], _[5], _[6])
                        await self.url_queue.put(item)
                        self.urls_enqueued.add(full_url)

            if self.args.full_scan and url.count('/') >= 2:
                await self.enqueue('/'.join(url.split('/')[:-2]) + '/')  # sub folder enqueue

            if url != '/' and not self.no_scripts:
                for script in self.user_scripts:
                    await self.url_queue.put((script, url))
            return True
        except Exception as e:
            await self.print_msg('[_enqueue.exception] %s' % str(e))
            return False

    #
    async def crawl(self, path, do_not_process_links=False):
        try:
            # increase body size to 200 KB
            request_headers = dict(conf.default_headers, Range='bytes=0-204800')
            status, headers, html_doc = await self.http_request(path, headers=request_headers)

            if path == '/':
                self.index_status, self.index_headers, self.index_html_doc = status, headers, html_doc
                if not self.index_has_reported:
                    self.index_has_reported = True
                    title = get_http_title(html_doc)
                    location = headers.get('Location', '')
                    server = headers.get('Server', '')
                    str_headers = ''
                    for key in self.index_headers:
                        # 减少非关键HTTP头的输出
                        if key.lower() in ['connection', 'content-encoding', 'content-security-policy',
                                           'date', 'p3p', 'x-ua-compatible', 'x-ua-compatible', 'cache-control',
                                           'x-xss-protection', 'transfer-encoding', 'last-modified', 'etag']:
                            continue
                        str_headers += '%s: %s\n' % (key, self.index_headers[key])
                    _ = {'status': status, 'url': clear_url(self.base_url), 'title': title, 'server': server,
                         'location': location, 'headers': str_headers}
                    await self.save_result('$Index', _)

                if self.fingerprint_check:
                    # 检查Web指纹
                    cms_name = fingerprint.get_cms_name('/^^^get^^^{}^^^', status, headers, html_doc)
                    if cms_name:
                        await self.save_result(
                            '$Fingerprint', cms_name,
                            msg='[Fingerprint] %s found %s' % (('%s%s' % (self.base_url, path)).rstrip('/'), cms_name))

                # 首页30x跳转，在第二次请求时，需要parse HTML， follow后获取新的HTML
                if not self.args.no_crawl and not do_not_process_links and status in [301, 302]:
                    resp = await self.conn_pool.get(self.base_url + '/',
                                                    headers=conf.default_headers, timeout=20)
                    location = resp.headers.get('Location', '')
                    if location.lower().startswith('http'):
                        scheme, netloc, _path, params, query, fragment = urlparse(location, 'http')
                        if netloc.find(self.host) < 0:   # different host, do not follow
                            location = ''
                        else:
                            location = _path + '?' + query
                    elif location.lower().startswith('/'):
                        pass
                    else:
                        location = '/' + location
                    if location:
                        url, depth = cal_depth(self, resp.headers.get('Location', ''))
                        if depth <= self.max_depth:
                            await self.enqueue(url)
                        # 避免处理错误，直接传入原始path，让httpx处理跳转URL，会重复，多1次请求
                        status, headers, html_doc = await self.http_request(path, headers=request_headers,
                                                                            follow_redirects=True)
                        # 再次检查Web指纹
                        cms_name = fingerprint.get_cms_name('/^^^get^^^{}^^^', status, headers, html_doc)
                        if cms_name:
                            await self.save_result(
                                '$Fingerprint', cms_name,
                                msg='[Fingerprint] %s found %s' % (
                                    ('%s%s' % (self.base_url, path)).rstrip('/'), cms_name))

            if not self.args.no_crawl and not do_not_process_links and html_doc:

                fav_url_found = False
                soup = BeautifulSoup(html_doc, "html.parser")
                for tag in ['link', 'script', 'a']:
                    for link in soup.find_all(tag):
                        origin_url = url = link.get('href', '').strip()
                        if not url:
                            origin_url = url = link.get('src', '').strip()
                        if url.startswith('..'):
                            continue
                        if not url.startswith('/') and url.find('//') < 0:   # relative path
                            url = path + url
                        url, depth = cal_depth(self, url)
                        # print(url, depth)
                        if depth <= self.max_depth:
                            await self.enqueue(url)
                        if self.fingerprint_check and tag == 'link' and str(link.get('rel', '')).find('icon') >= 0:
                            fav_url_found = True
                            fav_url, depth = cal_depth(self, link.get('href', '').strip())
                            if fav_url:    # 非当前域名的icon url，不会请求
                                await self.url_queue.put(('favicon', fav_url, ''))
                        # 解析js获取URL
                        if (path == '/' and tag == 'script' and (self.args.api or not self.args.fingerprint_only) and
                                origin_url not in self.js_urls):
                            self.js_urls.append(origin_url)
                            js_url, depth = cal_depth(self, origin_url)
                            if js_url:
                                if origin_url.lower().startswith('http') and origin_url.find('://') > 0:
                                    origin_url = origin_url.split('://')[1]
                                    if origin_url.find('/') > 0:
                                        origin_url = '/'.join(origin_url.split('/')[1:])
                                await self.url_queue.put(('js_file', origin_url, ''))

                if path == '/' and self.fingerprint_check and not fav_url_found:   # 尝试请求默认favicon，计算hash
                    await self.url_queue.put(('favicon', '/favicon.ico', ''))

                if path == '/' and self.fingerprint_check:
                    self.fingerprint_check = False    # this should only run once for each target
                    # 将CMS识别的其他请求，添加到队列
                    for key_name in fingerprint.rules.keys():
                        if key_name != '/^^^get^^^{}^^^':    # 首页已经默认请求过
                            await self.url_queue.put(('key_name', key_name, ''))

                ret = self.find_text(html_doc)
                if ret:
                    title = get_http_title(html_doc)
                    _ = {'status': status, 'url': '%s%s' % (self.base_url, path), 'title': title, 'vul_type': ret[1]}
                    await self.save_result('/', _)

        except Exception as e:
            await self.print_msg('[crawl Exception] %s %s %s' % (path, type(e), str(e)))

    async def load_all_urls_from_log_file(self):
        try:
            with open(self.log_file) as infile:
                for _line in infile.readlines():
                    _ = _line.strip().split()
                    if len(_) == 3 and (_[2].find('^^^200') > 0 or _[2].find('^^^403') > 0 or _[2].find('^^^302') > 0):
                        url, depth = cal_depth(self, _[1])
                        await self.enqueue(url)
        except Exception as e:
            await self.print_msg('[load_all_urls_from_log_file] %s' % str(e))

    def find_text(self, html_doc):
        for _text in self.text_to_find:
            if html_doc.find(_text) >= 0:
                return True, 'Found [%s]' % _text
        for _regex in self.regex_to_find:
            if _regex.search(html_doc):
                return True, 'Found Regex [%s]' % _regex.pattern
        return False

    def find_exclude_text(self, html_doc):
        for _text in self.text_to_exclude:
            if html_doc.find(_text) >= 0:
                return True
        for _regex in self.regex_to_exclude:
            if _regex.search(html_doc):
                return True
        return False

    async def is_url_valid(self, url, item):
        url_description, tag, status_to_match, content_type, content_type_no, root_only, vul_type = item
        status, headers, html_doc = await self.http_request(url)
        cur_content_type = headers.get('content-type', '')
        cur_content_length = headers.get('content-length', len(html_doc))

        if self.find_exclude_text(html_doc):  # excluded text found
            return False

        if 0 <= int(cur_content_length) <= 10:  # text too short
            return False

        if cur_content_type.find('image/') >= 0:  # exclude image
            return False

        if content_type != 'application/json' and cur_content_type.find('application/json') >= 0 and \
                not url.endswith('.json'):  # invalid json
            return False

        if content_type and cur_content_type.find(content_type) < 0 \
                or content_type_no and cur_content_type.find(content_type_no) >= 0:
            return False  # content type mismatch

        if tag and html_doc.find(tag) < 0:
            return False  # tag mismatch

        if self.find_text(html_doc):
            valid_item = True
        else:
            # status code check
            if status_to_match == 206 and status != 206:
                return False
            if status_to_match in (200, 206) and status in (200, 206):
                valid_item = True
            elif status_to_match and status != status_to_match:
                return False
            elif status in (403, 404) and status != status_to_match:
                return False
            else:
                valid_item = True

            if status == self._404_status and url != '/':
                len_doc = len(html_doc)
                len_sum = self.len_404_doc + len_doc
                if len_sum == 0 or (0.4 <= float(len_doc) / len_sum <= 0.6):
                    return False
        return valid_item

    async def save_result(self, prefix, item, msg=None):
        async with self.lock:
            if prefix not in self.results:
                self.results[prefix] = []
            if item not in self.results[prefix]:
                self.results[prefix].append(item)
                if msg:
                    await self.print_msg(msg)

    async def scan_worker(self):
        while True:
            if time.time() - self.start_time > self.time_out and not self.timeout_scan_aborted:
                self.timeout_scan_aborted = True
                clear_queue(self.url_queue)
                await self.print_msg('[ERROR] Timed out task: %s' % self.base_url)
                return
            try:
                item = self.url_queue.get_nowait()
            except Exception as e:
                return
            try:
                if len(item) == 3:
                    if item[0] == 'favicon':
                        resp = await self.conn_pool.get(self.base_url + item[1],
                                                        headers=conf.default_headers,
                                                        follow_redirects=False, timeout=20)
                        fav_hash = hashlib.md5(resp.content).hexdigest()
                        if fav_hash in fingerprint.fav_icons:
                            cms_name = fingerprint.fav_icons[fav_hash]
                            await self.save_result('$Fingerprint', cms_name,
                                                   msg='[Fingerprint] %s found %s' % (self.base_url, cms_name))

                    elif item[0] == 'key_name':
                        key_name = item[1]
                        req_item = fingerprint.requests_to_do[key_name]
                        if req_item[2]:
                            headers = copy.deepcopy(conf.default_headers)
                            headers.update(req_item[2])  # update headers
                        else:
                            headers = conf.default_headers
                        resp = None
                        if req_item[1].lower() == 'get':
                            resp = await self.conn_pool.get(self.base_url + req_item[0], headers=headers)
                        elif req_item[1].lower() == 'post':
                            data = req_item[3]
                            resp = await self.conn_pool.post(self.base_url + req_item[0], headers=headers, data=data)

                        if resp:
                            cms_name = fingerprint.get_cms_name(key_name, resp.status_code, resp.headers, resp.text)
                            if cms_name:
                                await self.save_result('$Fingerprint', cms_name,
                                                       '[Fingerprint] %s found %s' % (self.base_url, cms_name))

                    elif item[0] == 'js_file':
                        _path = item[1] if item[1].startswith('/') else '/' + item[1]
                        status, headers, js_doc = await self.http_request(_path)
                        if headers['content-type'].find('javascript') >= 0:
                            urls_regex, all_path_items, data_leak_found = await get_urls_in_js_async(
                                asyncio.get_event_loop(), js_doc, self.base_url + item[1], self.args.api, self)
                            # 目前并没有尝试请求匹配到的两组 疑似API接口，有误报，需要先优化正则，减少误报后，再添加
                            # 对于接口测试，这里应该是1个非常重要的检测点
                            if self.args.api:
                               self.urls_regex_found = self.urls_regex_found.union(urls_regex)

                            for item in all_path_items:
                                if type(item[2]) is str:
                                    if self.args.api:
                                        urls_regex.add(item[2])
                                    # await self.url_queue.put(('api_endpoint', item[2], ''))
                                    url, depth = cal_depth(self, item[2])
                                    if depth <= self.max_depth:
                                        await self.enqueue(url)

                            if data_leak_found:
                                for item in data_leak_found:
                                    _ = {'status': 200, 'url': self.base_url + _path,
                                         'title': '%s (%s)' % (item[1], item[2]), 'vul_type': 'JS Info Leak'}
                                    await self.save_result('/', _, '[JS Info Leak] %s : %s' % (_['url'], _['title']))

                    continue
                elif len(item) == 2:  # Script Scan
                    check_func = getattr(item[0], 'do_check')
                    # await self.print_msg('Begin %s %s' % (os.path.basename(item[0].__file__), item[1]))
                    await check_func(self, item[1])
                    # await self.print_msg('End %s %s' % (os.path.basename(item[0].__file__), item[1]))
                    continue
                else:
                    url_description, tag, status_to_match, content_type, content_type_no, root_only, vul_type = item
                    prefix = url_description['prefix']
                    url = url_description['full_url']

                    if url.find('{sub}') >= 0:
                        if not self.domain_sub:
                            continue
                        url = url.replace('{sub}', self.domain_sub)

            except Exception as e:
                await self.print_msg('[scan_worker.1] %s, %s, %s' % (str(e), self.base_url, item))
                # await self.print_msg(traceback.format_exc())
                continue
            if not item or not url:
                break

            try:
                valid_item = await self.is_url_valid(url, item)

                if valid_item:
                    _ = url.split('/')
                    _[-1] = 'fptest' + _[-1]
                    url_fp_test = '/'.join(_)  # add false positive test prefix
                    ret = await self.is_url_valid(url_fp_test, item)
                    if ret:
                        valid_item = False
                if valid_item:
                    status, headers, html_doc = await self.http_request(url)
                    title = get_http_title(html_doc)
                    _ = {'status': status, 'url': '%s%s' % (self.base_url, url), 'title': title, 'vul_type': vul_type}
                    await self.save_result(prefix, _)
            except Exception as e:
                await self.print_msg('[scan_worker.2][%s] %s, %s' % (url, str(e), item))
                # await self.print_msg(traceback.format_exc())

    async def scan(self, threads=6):
        try:
            all_threads = []
            for i in range(threads):
                t = self.scan_worker()
                all_threads.append(t)
            await asyncio.gather(*all_threads)

            for key in self.results.keys():
                # too many URLs found under this folder, deduplicate results
                if len(self.results[key]) > 10:
                    vul_type_count = {}
                    for item in copy.deepcopy(self.results[key]):
                        if item['vul_type'] not in vul_type_count:
                            vul_type_count[item['vul_type']] = 1
                        else:
                            vul_type_count[item['vul_type']] += 1
                            if vul_type_count[item['vul_type']] >= 3:
                                self.results[key].remove(item)
            return clear_url(self.base_url), self.results, self.urls_regex_found

        except Exception as e:
            await self.print_msg('[scan exception] %s' % str(e))
        finally:
            try:
                await self.conn_pool.aclose()
            except Exception as e:
                pass


async def scan_process():
    s = Scanner(args.timeout * 60)
    await s.init()
    while True:
        try:
            target = q_targets.get_nowait()
        except asyncio.queues.QueueEmpty as e:
            if conf.process_targets_done and q_targets.qsize() == 0:
                break
            else:
                await asyncio.sleep(0.1)
                continue

        if 'target' in target:
            ret = await s.init_from_url(target['target'])
        elif 'file' in target:
            ret = await s.init_from_log_file(target['file'])
        else:
            continue

        if ret:
            item = await s.scan(threads=args.t)
            if item[1]:
                await q_results.put(copy.deepcopy(item))


async def add_target(target, is_neighbor=False):
    if is_neighbor:
        target['no_scripts'] = 1    # 邻居IP，不启用插件. Bug fixed: 2024/05/03
    if args.debug:
        await q_results.put('New target: %s' % target)
    await q_targets.put({'target': target})
    if args.save_ports and target['ports_open']:
        conf.ports_saved_to_file = True
        if not args.ports_file:
            args.ports_file = open(args.save_ports, 'w')
        for port in target['ports_open']:
            args.ports_file.write('%s:%s\n' % (target['host'], port))
        args.ports_file.flush()
    conf.tasks_count += 1


def is_intranet(ip):
    try:
        ret = ip.split('.')
        if len(ret) != 4:
            return True
        if ret[0] == '10':
            return True
        if ret[0] == '172' and 16 <= int(ret[1]) <= 31:
            return True
        if ret[0] == '192' and ret[1] == '168':
            return True
        return False
    except Exception as e:
        return False


resolver = dns.asyncresolver.Resolver()


async def domain_lookup_check(queue_targets_origin, processed_targets, queue_targets):
    while True:
        try:
            url = queue_targets_origin.get_nowait()
        except asyncio.queues.QueueEmpty as e:
            break
        # scheme netloc path
        if url.find('://') < 0:
            netloc = url[:url.find('/')] if url.find('/') > 0 else url
        else:
            scheme, netloc, path, params, query, fragment = urlparse(url, 'http')

        # host port
        host = netloc.split(':')[0] if netloc.find(':') >= 0 else netloc

        if is_ip_addr(host):
            processed_targets.append(host)
            if args.skip_intranet and is_intranet(host):
                await q_results.put('Private IP target skipped: %s [%s]' % (url, host))
            else:
                await queue_targets.put((url, 0, host))
        else:
            for i in range(5):
                try:
                    answers = await resolver.resolve(host, "A")
                    processed_targets.append(answers[0].address)
                    if args.skip_intranet and is_intranet(answers[0].address):
                        await q_results.put('Private IP target skipped: %s [%s]' % (url, answers[0].address))
                    else:
                        await queue_targets.put((url, 0, answers[0].address))
                    break
                except dns.resolver.NXDOMAIN as e:
                    await q_results.put('No such domain: %s' % host)
                    break
                except Exception as e:
                    if i == 4:    # Failed after 4 retries
                        await q_results.put('Domain lookup failed [%s]: %s' % (e.__class__.__name__, host))


async def do_port_scan_check(queue_targets):
    """
    检测目标的端口是否开放，输入的目标是URL，也可能是网段下的相邻IP
    """
    while True:
        try:
            url, is_neighbor, ip_addr = queue_targets.get_nowait()   # is_neighbor = 1 为相邻网段的IP，优先级降低
        except asyncio.queues.QueueEmpty as e:
            break
        try:
            # scheme netloc path
            if url.find('://') < 0:
                scheme = 'unknown'
                netloc = url[:url.find('/')] if url.find('/') > 0 else url
                path = ''
            else:
                scheme, netloc, path, params, query, fragment = urlparse(url, 'http')

            # host port
            if netloc.find(':') >= 0:
                _ = netloc.split(':')
                host = _[0]
                try:
                    port = int(_[1])
                except:
                    port = None
            else:
                host = netloc
                port = None

            if scheme == 'https' and port is None:
                port = 443
            elif scheme == 'http' and port is None:
                port = 80

            if scheme == 'unknown':
                if port == 80:
                    scheme = 'http'
                if port == 443:
                    scheme = 'https'

            ports_open = set()
            ports_closed = set()

            # 插件不依赖HTTP连接池， 且仅启用插件扫描， 则不需要检查80/443端口的HTTP服务， 直接扫描 require_ports
            if args.scripts_only and args.require_no_http:
                ports_open, ports_closed = await scan_given_ports(ip_addr, args.require_ports, ports_open, ports_closed)
                target = {'scheme': scheme, 'host': host, 'port': port, 'path': path,
                          'has_http': False, 'ports_open': ports_open, 'ports_closed': ports_closed}
                await add_target(target)    # 在只扫插件的情况下，相邻IP也需要启用
                continue

            if port:
                # 指定了 标准端口 或 非标准端口
                has_http = await is_port_open(ip_addr, port)
                if has_http:
                    ports_open.add(port)
                else:
                    ports_closed.add(port)
                if not args.no_scripts:
                    ports_open, ports_closed = \
                        await scan_given_ports(ip_addr, args.require_ports, ports_open, ports_closed)

                target = {'scheme': scheme, 'host': host, 'port': port, 'path': path, 'has_http': has_http,
                          'ports_open': ports_open, 'ports_closed': ports_closed}
                await add_target(target)

            else:
                # 只有域名和IP情况下， 扫默认端口
                port_open_80 = await is_port_open(ip_addr, 80)
                port_open_443 = await is_port_open(ip_addr, 443)

                if port_open_80:
                    ports_open.add(80)
                else:
                    ports_closed.add(80)
                if port_open_443:
                    ports_open.add(443)
                else:
                    ports_closed.add(443)
                if not args.no_scripts:
                    ports_open, ports_closed = \
                        await scan_given_ports(ip_addr, args.require_ports, ports_open, ports_closed)

                if port_open_80 and port_open_443:
                    target = {'scheme': 'https', 'host': host, 'port': 443, 'path': path,
                              'has_http': True, 'ports_open': ports_open, 'ports_closed': ports_closed}
                    await add_target(target, is_neighbor)
                    # 排除 301 HTTP 跳转 HTTPS的目标
                    async with httpx.AsyncClient() as client:
                        r = await client.get('http://%s' % host, follow_redirects=False, timeout=20)
                        if r and not \
                                (r.status_code == 301 and r.headers.get('Location', '').lower().startswith('https')):
                            target = {'scheme': 'http', 'host': host, 'port': 80, 'path': path,
                                      'has_http': True, 'no_scripts': 1,
                                      'ports_open': ports_open, 'ports_closed': ports_closed}
                            await add_target(target)

                elif port_open_443:
                    target = {'scheme': 'https', 'host': host, 'port': 443, 'path': path,
                              'has_http': True, 'ports_open': ports_open, 'ports_closed': ports_closed}
                    # 即使指定的目标，允许插件扫描，邻居也将不启用，节省扫描时间
                    await add_target(target, is_neighbor)
                elif port_open_80:
                    target = {'scheme': 'http', 'host': host, 'port': 80, 'path': path,
                              'has_http': True, 'ports_open': ports_open, 'ports_closed': ports_closed}
                    await add_target(target, is_neighbor)
                elif args.no_scripts:
                    # 80 443 端口不开放， 禁用插件扫描
                    await q_results.put('No ports open: %s' % host)
                elif not is_neighbor or args.scripts_only:
                    # 直接输入目标 或者 对相邻IP应用插件
                    # 80 443 未开放，此时只能检测其他端口的漏洞
                    # 如果没有任何开放的端口，直接跳过该目标
                    if ports_open:
                        target = {'scheme': 'http', 'host': host, 'port': 80, 'path': path,
                                  'has_http': False, 'ports_open': ports_open, 'ports_closed': ports_closed}
                        await add_target(target)
                    else:
                        await q_results.put('[Warning] Target has no open ports: %s' % url)
        except (httpx.RequestError, httpx.HTTPStatusError) as e:
            pass
        except Exception as e:
            # import traceback
            # await q_results.put(traceback.format_exc())
            await q_results.put('[port_scan_check.exception] URL is %s, %s' % (url, str(e)))


async def port_scan_check(queue_targets):
    threads = [do_port_scan_check(queue_targets) for _ in range(250)]
    await asyncio.gather(*threads)


async def prepare_targets(target_list):
    """
    Process URL / IP / Domain, port scan
    处理域名、IP，扫描目标端口80 443等端口是否开放
    """
    queue_targets_origin = asyncio.Queue()
    for target in target_list:
        if target.strip() and len(target) > 5:
            # work with https://github.com/lijiejie/subDomainsBrute
            # Delimiter should be ","
            hosts = target.replace(',', ' ').strip().split()
            await queue_targets_origin.put(hosts[0])

    processed_targets = []
    # 将域名解析和端口扫描拆分，可节省约2s.更简单的做法, 可以将DNS解析和端口扫描合并为一个函数，但会损失 2s
    await q_results.put('Domain lookup start.')
    queue_targets = asyncio.Queue()
    # Be careful: 当 DNS查询并发过高时，在家庭网络下会出现较多超时
    threads = [domain_lookup_check(queue_targets_origin, processed_targets, queue_targets) for _ in range(50)]
    await asyncio.gather(*threads)

    if args.network != 32:
        await q_results.put('Process sub network start.')
        num_entered_queue = 0
        for ip in processed_targets:
            if ip.find('/') > 0:    # 子网本身已经处理过
                continue
            _network = u'%s/%s' % ('.'.join(ip.split('.')[:3]), args.network)
            if _network in processed_targets:
                continue
            processed_targets.append(_network)

            if args.network >= 20:
                sub_nets = [ipaddress.IPv4Network(u'%s/%s' % (ip, args.network), strict=False).hosts()]
            else:
                sub_nets = ipaddress.IPv4Network(u'%s/%s' % (ip, args.network), strict=False).subnets(new_prefix=22)
            for sub_net in sub_nets:
                if sub_net in processed_targets:
                    continue
                if type(sub_net) is ipaddress.IPv4Network:    # add network only
                    processed_targets.append(str(sub_net))
                for _ip in sub_net:
                    _ip = str(_ip)
                    if _ip not in processed_targets:
                        await queue_targets.put((_ip, 1, _ip))
                        num_entered_queue += 1
                if num_entered_queue > 65535:    # 队列不宜太长，如果超过一个B段，分多次处理
                    await port_scan_check(queue_targets)
                    num_entered_queue = 0
    if queue_targets.qsize() > 0:    # 还有剩余未处理目标
        await port_scan_check(queue_targets)
    # save ports data
    if args.save_ports and args.ports_file:
        args.ports_file.close()

    conf.process_targets_done = True
    await q_results.put('* Targets DNS resolve and port scan all done.')


async def main():
    for input_file in args.input_files:
        if args.host:
            target_list = args.host
            # Targets input via commandline args, create double processes at most
            if args.network == 32 and len(target_list) * 2 < args.p:
                args.p = len(target_list) * 2
        elif args.f or args.d:
            with codecs.open(input_file, encoding='utf-8') as inFile:
                target_list = inFile.readlines()
                # Targets input via file, create double processes at most
                if args.network == 32 and len(target_list) * 2 < args.p:
                    args.p = len(target_list) * 2
        try:
            clear_queue(q_results)
            clear_queue(q_targets)
            # save report thread
            asyncio.create_task(save_report(args, q_results, input_file))

            conf.process_targets_done = False
            start_time = time.time()

            if args.crawler:
                # 爬虫URL导入，在3.0版本后，还未经测试，仅保留了原逻辑。 待测试
                input_files = glob.glob(args.crawler + '/*.log')
                for _file in input_files:
                    await q_targets.put({'file': _file})
                    conf.tasks_count += 1
                if conf.tasks_count < args.p:
                    args.p = conf.tasks_count
                conf.process_targets_done = True
            else:
                conf.tasks_count = 0
                asyncio.create_task(prepare_targets(target_list))

            all_process = [scan_process() for _ in range(args.p)]
            await q_results.put('%s scan process started' % args.p)
            await asyncio.gather(*all_process)

            cost_time = time.time() - start_time
            cost_min = int(cost_time / 60)
            cost_min = '%s min ' % cost_min if cost_min > 0 else ''
            cost_seconds = '%.1f' % (cost_time % 60)
            await q_results.put('Scanned %s targets in %s%s seconds' % (conf.tasks_count, cost_min, cost_seconds))
        except KeyboardInterrupt as e:
            conf.stop_me = True
            await q_results.put('Scan aborted by user')
            if conf.output_file_name:
                await q_results.put('If you are interested, partial report is: %s' % conf.output_file_name)
            exit(-1)
        except Exception as e:
            traceback.print_exc()
            await q_results.put('[main.exception] %s %s' % (type(e), str(e)))

        conf.stop_me = True
        await asyncio.sleep(3.0)   # report 需要一些时间写入和唤起浏览器


if __name__ == '__main__':
    args = parse_args()
    print('* BBScan %s  https://github.com/lijiejie/BBScan *' % conf.version)
    if args.no_scripts:
        print('* Scripts scan was disabled')
    if args.require_ports:
        print('* Scripts scan port check: %s' % ','.join([str(x) for x in args.require_ports]))
    if sys.version_info.major >= 3 and sys.version_info.minor >= 10:
        loop = asyncio.new_event_loop()
    else:
        loop = asyncio.get_event_loop()
    q_targets = asyncio.Queue()    # targets Queue
    q_results = asyncio.Queue()    # results Queue
    loop.run_until_complete(main())



# -*- encoding: utf-8 -*-
# Parse Javascript file and try to find API interfaces and possible data leaks


import time
from pyjsparser import parse
import copy
import re


async def get_urls_in_js_async(loop, code, js_url, with_regex, self):
    start_time = time.time()
    all_path_items, data_leak_found, error = await loop.run_in_executor(None, get_urls_in_js, code)
    if self.args.debug:
        await self.print_msg('Parse Js: %s, cost %.2f seconds' % (js_url, time.time() - start_time))
        if error:
            await self.print_msg('Js parse failed, url: %s %s' % (js_url, error))

    # key关键词误报临时处理，待优化
    for item in copy.deepcopy(data_leak_found):
        if item[1] == 'key' and (
                code.find('this.%s(' % item[2]) > 0 or code.find('%s:function(' % item[2]) > 0 or
                code.find('{key:"%s",value:function' % item[2]) > 0 or
                code.find('{key:"%s",fn:function' % item[2]) > 0 or
                code.find('{key:"%s",get:function' % item[2]) > 0 or
                code.find('{key:"%s",set:function' % item[2]) > 0):
            # await self.print_msg('js parse, remove %s' % item)
            data_leak_found.remove(item)
    if with_regex:
        urls = extract_urls_regex(code)
    else:
        urls = set([])
    return urls, all_path_items, data_leak_found


def is_interested_var_name(var_name):
    # 查找 key / token / secret 泄露
    var_name = var_name.lower()
    for word in ['key', 'token', 'secret', 'password']:
        if var_name == word:
            return True
        if var_name.endswith('_' + word) or var_name.find('_' + word + '_') >= 0 or var_name.endswith(word + '_'):
            return True

    return False


def extract_urls_regex(code):
    # Regex from: https://github.com/Threezh1/JSFinder/blob/master/JSFinder.py
    regex_str = r"""

      (?:"|')                               # Start newline delimiter

      (
        ((?:[a-zA-Z]{1,10}://|//)           # Match a scheme [a-Z]*1-10 or //
        [^"'/]{1,}\.                        # Match a domainname (any character + dot)
        [a-zA-Z]{2,}[^"']{0,})              # The domainextension and/or path

        |

        ((?:/|\.\./|\./)                    # Start with /,../,./
        [^"'><,;| *()(%%$^/\\\[\]]          # Next character can't be...
        [^"'><,;|()]{1,})                   # Rest of the characters can't be

        |

        ([a-zA-Z0-9_\-/]{1,}/               # Relative endpoint with /
        [a-zA-Z0-9_\-/.]{1,}                # Resource name
        \.(?:[a-zA-Z]{1,4}|action)          # Rest + extension (length 1-4 or action)
        (?:[\?|#][^"|']{0,}|))              # ? or # mark with parameters

        |

        ([a-zA-Z0-9_\-/]{1,}/               # REST API (no extension) with /
        [a-zA-Z0-9_\-/]{3,}                 # Proper REST endpoints usually have 3+ chars
        (?:[\?|#][^"|']{0,}|))              # ? or # mark with parameters

        |

        ([a-zA-Z0-9_\-]{1,}                 # filename
        \.(?:php|asp|aspx|jsp|json|
             action|html|js|txt|xml)        # . + extension
        (?:[\?|#][^"|']{0,}|))              # ? or # mark with parameters

      )

      (?:"|')                               # End newline delimiter

    """
    pattern = re.compile(regex_str, re.VERBOSE)
    result = re.finditer(pattern, code)
    if not result:
        return None
    ret = set()
    # 初步过滤误报，待优化正则表达式
    for match in result:
        item = match.group().strip('"').strip("'")
        if (item.count('/') == 1 and item.startswith('application/') or item.startswith('text/') or
                item.startswith('audio/') or item.startswith('multipart/') or item.startswith('video/')):
            continue
        if url_in_known_sites(item):
            continue
        if item.count('/') == 2 and all(x.lower() in ['yyyy', 'yy', 'mm', 'm', 'dd', 'd'] for x in item.split('/')):
            continue
        if item in ['/./']:
            continue
        if any(item.endswith(x) for x in ['.vue', '.jpg', '.png', '.svg', '.gif']):
            continue
        ret.add(item)
    return ret


def url_in_known_sites(url):
    for site in ['registry.npmjs.org', 'www.w3.org', 'github.com', 'registry.npmjs.org', 'ant.design']:
        if url.startswith('http://' + site) or url.startswith('https://' + site):
            return True


def get_expression_value(exp):
    value = ''
    try:
        if exp['type'] == 'Literal':    # 常量
            return exp['value']
        elif exp['type'] == 'BinaryExpression':    # 二元操作符
            for key in exp:
                if 'type' in exp[key] and exp[key]['type'] == 'Literal':
                    # select a str which has most slash /
                    if type(exp[key]['value']) is str and exp[key]['value'].count('/') > value.count('/'):
                        value = exp[key]['value']
        elif exp['type'] == 'ObjectExpression':    # 对象, 字典
            for property in exp['properties']:
                tmp = get_expression_value(property['value'])
                if type(tmp) is str and tmp.count('/') > value.count('/'):
                    value = tmp
    except Exception as e:
        print('get_expression_value.exception: %s' % str(e))
    return value


def get_urls_in_js(code):
    try:
        r = parse(code)
    except Exception as e:
        return [], [], str(e)
    traverse_list = [r]    # 从root 开始遍历
    all_path_items = []
    data_leak_found = []
    counter = 0
    while True:
        counter += 1
        if len(traverse_list) == 0 or counter > 10000:
            return all_path_items, data_leak_found, None
        item = traverse_list.pop()
        if type(item) is dict:
            for key in item:
                if type(item[key]) in (list, dict):
                    traverse_list.append(item[key])

                # 变量声明
                if key == 'type' and item['type'] == 'VariableDeclarator':
                    if 'id' in item:
                        var_name = item['id']['name'].lower()
                        if var_name.find('path') >= 0 or var_name.find('uri') >= 0 or var_name.find('url') >= 0:
                            if 'init' not in item or not item['init'] or 'value' not in item['init']:
                                continue
                            path = item['init']['value']
                            if path:
                                all_path_items.append(['a_value', var_name, path])
                        if is_interested_var_name(var_name):
                            if 'init' not in item or not item['init'] or 'value' not in item['init']:
                                continue
                            value = item['init']['value']
                            if value:
                                data_leak_found.append(['a_value', var_name, value])

                if key == 'key':
                    if 'name' not in item['key']:
                        continue
                    var_name = item['key']['name'].lower()
                    if var_name.find('path') >= 0 or var_name.find('uri') >= 0 or var_name.find('url') >= 0:
                        if item['value']['type'] == 'Literal':
                            if item['value']['value']:
                                all_path_items.append(['a_value', var_name, item['value']['value']])
                        elif item['value']['type'] == 'BinaryExpression':
                                val = get_expression_value(item['value'])
                                if val:
                                    all_path_items.append(['a_value',  var_name, get_expression_value(item['value'])])
                    if is_interested_var_name(var_name):
                        if item['value']['type'] == 'Literal':
                            if item['value']['value']:
                                data_leak_found.append(['a_value', var_name, item['value']['value']])
                        elif item['value']['type'] == 'BinaryExpression':
                                val = get_expression_value(item['value'])
                                data_leak_found.append(['a_value', var_name, val])

                # 函数调用
                if key == 'callee':
                    if 'property' not in item['callee'] or 'name' not in item['callee']['property']:
                        continue
                    func_name = item['callee']['property']['name']
                    # 处理ajax请求的参数
                    if func_name.lower() in ['get', 'post', 'put', 'ajax']:
                        if item['arguments']:
                            val = get_expression_value(item['arguments'][0])
                            if val:
                                all_path_items.append(['a_function', func_name, val])
                    if is_interested_var_name(func_name):
                        if item['arguments']:
                            val = get_expression_value(item['arguments'][0])
                            if val:
                                data_leak_found.append(['a_function', func_name, val])

        elif type(item) is list:
            for child_item in item:
                if type(child_item) in (list, dict):
                    traverse_list.append(child_item)



# Global Variables share among modules

stop_me = False

output_file_name = None

user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) ' \
             'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.0.0 Safari/537.36'

default_headers = {'User-Agent': user_agent, 'Range': 'bytes=0-102400'}

ports_saved_to_file = False

process_targets_done = False   # scan coroutine will wait until process_targets_done

tasks_count = 0    # task counter

version = 'v3.0'



# Wordpress

from lib.common import save_script_result


async def do_check(self, url):
    if url == '/' and self.conn_pool:
        if self.index_html_doc.find('/wp-content/themes/') >= 0:
            url_lst = ['/wp-config.php.inc',
                       '/wp-config.inc',
                       '/wp-config.bak',
                       '/wp-config.php~',
                       '/.wp-config.php.swp',
                       '/wp-config.php.bak']
            for _url in url_lst:
                status, headers, html_doc = await self.http_request(_url)
                if status == 200 or status == 206:
                    if html_doc.find('<?php') >= 0:
                        await save_script_result(self, status, self.base_url + _url, '', 'WordPress Backup File Found')



# -*- encoding: utf-8 -*-
# Indentify web app fingerprints: framework, programming languages, web server, CMS,
# middle-ware, open source software or commercial product etc
# Rules copy from https://github.com/0x727/FingerprintHub

import hashlib
import os
import json
import codecs
import httpx

cur_dir = os.path.dirname(os.path.abspath(__file__))
rule_dir = os.path.join(cur_dir, '../rules/web_fingerprint_v3.json')


class Fingerprint(object):
    def __init__(self):
        self.fav_icons = {}
        self.requests_to_do = {}
        self.rules = {}

        with codecs.open(rule_dir, encoding='utf-8') as f:
            doc = json.loads(f.read())

            for rule in doc:
                # 处理fav hash
                if rule['favicon_hash']:
                    for _hash in rule['favicon_hash']:
                        if _hash:
                            self.fav_icons[_hash] = rule['name']

                key = '^^^'.join([rule['path'], rule['request_method'],
                                  str(rule['request_headers']), rule['request_data']])
                self.requests_to_do[key] = [rule['path'], rule['request_method'],
                                            rule['request_headers'], rule['request_data']]
                if key not in self.rules:
                    self.rules[key] = []
                self.rules[key].append(rule)

    def get_cms_name_via_icon(self, favicon_hash):
        if favicon_hash in self.fav_icons:
            return self.fav_icons[favicon_hash]
        else:
            return

    def get_cms_name(self, key_name, status_code, headers, text, favicon_hash=None):
        cms_names = []
        for rule in self.rules[key_name]:
            if rule['status_code'] != 0:
                # 200 和 206 单独检查
                if rule['status_code'] in [200, 206] and status_code in [200, 206]:
                    pass
                else:
                    if rule['status_code'] != status_code:  # code mismatch
                        continue
            mismatch = False
            if rule['headers']:
                for header_name in rule['headers']:
                    if rule['headers'][header_name] == '*' and header_name in headers:
                        continue
                    if headers.get(header_name, '').find(rule['headers'][header_name]) < 0:
                        mismatch = True
                        break
                if mismatch:
                    continue
            if rule['keyword']:
                for word in rule['keyword']:
                    if text.lower().find(word) < 0 and text.find(word) < 0:
                        mismatch = True
                        break
                if mismatch:
                    continue
            if rule['favicon_hash'] and favicon_hash != rule['favicon_hash']:
                continue
            if rule['name'] not in cms_names:
                cms_names.append(rule['name'])
        return cms_names


if __name__ == '__main__':
    from config import default_headers
    import copy
    f = Fingerprint()

    client = httpx.Client()
    data = client.get('https://demo.jumpserver.org/static/img/facio.ico').read()
    fav_hash = hashlib.md5(data).hexdigest()
    if fav_hash in f.fav_icons:
        print('From fav hash:', f.fav_icons[fav_hash])

    url = 'http://example.com/'

    for key_name in f.rules:
        item = f.requests_to_do[key_name]
        print(key_name)
        print()

        if item[2]:
            headers = copy.deepcopy(default_headers)
            headers.update(item[2])    # update headers
        else:
            headers = default_headers

        resp = None
        if item[1].lower() == 'get':
            resp = client.get(url.rstrip('/') + item[0], headers=headers)
        elif item[1].lower() == 'post':
            data = item[3]
            resp = client.post(url.rstrip('/') + item[0], headers=headers, data=item[3])
        else:
            raise Exception('invalid method')

        if resp:
            cms_name = f.get_cms_name(key_name, resp.status_code, resp.headers, resp.text)
            if cms_name:
                print('cms name is:', cms_name)




from lib.common import save_script_result


async def do_check(self, url):
    if url == '/' and self.conn_pool:
        folders = ['']
        for log_folder in ['false_positive_log', 'log', 'logs']:
            status, headers, html_doc = await self.http_request('/' + log_folder)

            if status in (301, 302):
                location = headers.get('location', '')
                if location.startswith(self.base_url + '/' + log_folder + '/') or \
                        location.startswith('/' + log_folder + '/'):
                    if log_folder == 'false_positive_log':    # false positive check
                        break
                    folders.append(log_folder)
                    await self.enqueue(log_folder)
                    await self.crawl('/' + log_folder + '/')

            if status == 206 and self._404_status != 206:
                await save_script_result(self, status, self.base_url + '/' + log_folder, '',
                                   'Log File Found')

        url_lst = ['false_positive_access.log', 'access.log', 'www.log', 'error.log', 'log.log', 'sql.log',
                   'log/false_positive_access.log','log/access.log',
                   'errors.log', 'debug.log', 'db.log', 'install.log',
                   'server.log', 'sqlnet.log', 'WS_FTP.log', 'database.log', 'data.log', 'app.log',
                   'false_positive_log.tar.gz', 'log.tar.gz', 'log.rar', 'log.zip',
                   'log.tgz', 'log.tar.bz2', 'log.7z']

        for log_folder in folders:
            for _url in url_lst:
                url_prefix = '/' + log_folder if log_folder else ''
                status, headers, html_doc = await self.http_request(url_prefix + '/' + _url)
                # print '/' + log_folder + '/' + _url
                if status in [206, 200] and headers.get('content-type', '').find('application/') >= 0:
                    if _url == url_lst[0] or _url in ['false_positive_access.log',
                                                      'false_positive_log.tar.gz', 'log/false_positive_access.log']:
                        break
                    # false positive check
                    ret = _url.split('/')
                    ret[-1] = 'fptest' + ret[-1]
                    status, headers, html_doc = await self.http_request(url_prefix + '/' + '/'.join(ret))
                    # print '/' + log_folder + '/' + _url
                    if status in [206, 200] and headers.get('content-type', '').find('application/') >= 0:
                        pass
                    elif self.find_exclude_text(html_doc):
                        pass
                    else:
                        await save_script_result(self, status, self.base_url + url_prefix + '/' + _url,
                                                 '', 'Log File')

        for log_folder in folders:
            for _url in ['log_false_positive.txt', 'log.txt', 'logs.txt']:
                url_prefix = '/' + log_folder if log_folder else ''
                status, headers, html_doc = await self.http_request(url_prefix + '/' + _url)
                # print '/' + log_folder + '/' + _url
                if status == 206 and headers.get('content-type', '').find('text/plain') >= 0:
                    if _url == 'log_false_positive.txt':
                        break
                    await save_script_result(self, status, self.base_url + url_prefix + '/' + _url,
                                             '', 'Log File')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#

from urllib.parse import urlparse
import re
import asyncio
import platform
import socket
import dns.asyncresolver
import time


def get_dns_resolver():
    resolver = dns.asyncresolver.Resolver()
    for server in ['114.114.114.114', '180.76.76.76', '8.8.8.8']:  # Add public DNS Server
        if server not in resolver.nameservers:
            resolver.nameservers.append(server)
    return resolver


if platform.system() == 'Windows':
    try:
        def _call_connection_lost(self, exc):
            try:
                self._protocol.connection_lost(exc)
            finally:
                if hasattr(self._sock, 'shutdown'):
                    try:
                        if self._sock.fileno() != -1:
                            self._sock.shutdown(socket.SHUT_RDWR)
                    except Exception as e:
                        pass
                self._sock.close()
                self._sock = None
                server = self._server
                if server is not None:
                    server._detach()
                    self._server = None

        asyncio.proactor_events._ProactorBasePipeTransport._call_connection_lost = _call_connection_lost
    except Exception as e:
        pass


def is_ip_addr(s):
    pattern_ip = re.compile(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$')
    ret = pattern_ip.search(s)
    return True if ret else False


def clear_queue(this_queue):
    try:
        while True:
            this_queue.get_nowait()
    except Exception as e:
        return


def parse_url(url):
    _ = urlparse(url, 'http')
    if not _.netloc:
        _ = urlparse('https://' + url, 'http')
    return _.scheme, _.netloc, _.path if _.path else '/'


# calculate depth of a given URL, return tuple (url, depth)
def cal_depth(self, url):
    if url.find('#') >= 0:
        url = url[:url.find('#')]  # cut off fragment
    if url.find('?') >= 0:
        url = url[:url.find('?')]  # cut off query string

    while url.find('/./') >= 0:
        url = url.replace('/./', '/')

    if url.startswith('//'):
        return '', 10000  # //www.baidu.com/index.php

    if not urlparse(url, 'http').scheme.startswith('http'):
        return '', 10000  # no HTTP protocol

    if url.lower().startswith('http'):
        _ = urlparse(url, 'http')
        if _.netloc == self.host or _.netloc == '%s:%s' % (self.host, self.port):  # same hostname
            url = _.path
        else:
            return '', 10000  # not the same hostname

    while url.find('//') >= 0:
        url = url.replace('//', '/')

    if not url:
        return '/', 1  # http://www.example.com

    if url[0] != '/':
        url = '/' + url

    url = url[: url.rfind('/') + 1]

    if url.split('/')[-2].find('.') > 0:
        url = '/'.join(url.split('/')[:-2]) + '/'

    depth = url.count('/')
    # print('cal_depth', url, depth)
    return url, depth


async def save_script_result(self, status, url, title, vul_type=''):
    async with self.lock:
        # print '[+] [%s] %s' % (status, url)
        if url not in self.results:
            self.results[url] = []
        _ = {'status': status, 'url': url, 'title': title, 'vul_type': vul_type}
        self.results[url].append(_)


def get_domain_sub(host):
    if re.search(r'\d+\.\d+\.\d+\.\d+', host.split(':')[0]):
        return ''
    else:
        return host.split('.')[0]


def escape(html):
    return html.replace('&', '&amp;').\
        replace('<', '&lt;').replace('>', '&gt;').\
        replace('"', '&quot;').replace("'", '&#39;')


sem = asyncio.Semaphore(100)

resolver = get_dns_resolver()


async def is_port_open(host, port):
    if not port:
        return True

    try:
        async with sem:
            start_time = time.time()
            if not is_ip_addr(host):
                answers = await resolver.resolve(host, "A")
                host = answers[0].address

            fut = asyncio.open_connection(host, int(port))
            reader, writer = await asyncio.wait_for(fut, timeout=10)
            writer.close()
            try:
                await writer.wait_closed()    # application data after close notify (_ssl.c:2730)
            except Exception as e:
                print('is_port_open.wait_closed.exception:', type(e))
            return True
    except (asyncio.exceptions.TimeoutError, ConnectionRefusedError) as e:
        pass
    except Exception as e:
        print('is_port_open.exception:', e.__class__.__name__, str(e), host, port,
              'elapsed %.2f seconds' % (time.time() - start_time))
        return False


def get_http_title(html_doc):
    if not html_doc:
        return ''
    m = re.search('<title>(.*?)</title>', html_doc, re.IGNORECASE)
    return m.group(1) if m else ''


def clear_url(base_url):
    if base_url.startswith('unknown://'):
        base_url = base_url[len('unknown://'):]
    if base_url.endswith(':None'):
        base_url = base_url[:-len(':None')]
    return base_url


async def scan_given_ports(host, ports, confirmed_open, confirmed_closed):
    checked_ports = confirmed_open.union(confirmed_closed)
    ports_open = set()
    ports_closed = set()

    scanning_ports = []
    threads = []
    for port in ports:
        if port not in checked_ports:   # 不重复检测已确认端口
            scanning_ports.append(port)
            threads.append(is_port_open(host, port))
    ret = await asyncio.gather(*threads)
    for i in range(len(threads)):
        if ret[i]:
            ports_open.add(scanning_ports[i])
        else:
            ports_closed.add(scanning_ports[i])

    return ports_open.union(confirmed_open), ports_closed.union(confirmed_closed)


async def test():
    r = await is_port_open('www.baidu.com', 80)
    print(r)


def run_test_is_port_open():
    loop = asyncio.get_event_loop()
    task = loop.create_task(test())
    loop.run_until_complete(task)


if __name__ == '__main__':
    run_test_is_port_open()



# Outlook Web APP

import httpx
from lib.common import save_script_result


async def do_check(self, url):
    if url == '/' and self.conn_pool:
        if self.index_status == 302 and self.index_headers.get('location', '').lower() == 'https://%s/owa' % self.host:
            await save_script_result(self, 302, 'https://%s' % self.host, 'OutLook Web APP Found')
            return

        status, headers, html_doc = await self.http_request('/ews/')

        if status == 302:
            redirect_url = headers.get('location', '')
            if redirect_url == 'https://%shttp://%s/ews/' % (self.host, self.host):
                await save_script_result(self, 302, 'https://%s' % self.host, 'OutLook Web APP Found')
                return
            if redirect_url == 'https://%s/ews/' % self.host:
                try:
                    async with httpx.AsyncClient() as client:
                        r = await client.head('/ews')
                        if r.status_code == 401:
                            await save_script_result(self, 401, redirect_url, 'OutLook Web APP Found')
                except Exception as e:
                    pass
                return

        elif status == 401:
            if headers.get('Server', '').find('Microsoft-IIS') >= 0:
                await save_script_result(self, 401, self.base_url + '/ews/', 'OutLook Web APP Found')
                return



#!/usr/bin/python
# -*- encoding: utf-8 -*-

import motor.motor_asyncio
from lib.common import save_script_result


ports_to_check = 27017    # 默认扫描端口


async def do_check(self, url):
    if url != '/':
        return
    port = 27017
    if self.scheme == 'mongodb' and self.port != 27017:    # 非标准端口
        port = self.port
    elif 27017 not in self.ports_open:
        return
    try:
        client = motor.motor_asyncio.AsyncIOMotorClient(self.host, port)
        ret = await client.list_database_names()
        detail = "%s MongoDB Unauthorized Access : %s" % (self.host, ",".join(ret))
        await save_script_result(self, '', 'mongodb://%s:%s' % (self.host, port), detail)
    except Exception as e:
        pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# ports_to_check 设置为 想要扫描的1个或多个端口
# python BBScan.py --scripts-only --script port_scan --host www.baidu.com --network 16 --save-ports ports_80.txt

ports_to_check = [80]


def do_check(self, url):
    pass


from lib.common import save_script_result


async def do_check(self, url):
    if url == '/' and self.conn_pool:
        if self.index_status == 301 and self.index_headers.get('location', '').find('forum.php') >= 0 or \
                str(self.index_headers).find('_saltkey=') > 0:

            url_lst = ['/config/config_ucenter.php.bak',
                       '/config/.config_ucenter.php.swp',
                       '/config/.config_global.php.swp',
                       '/config/config_global.php.1',
                       '/uc_server/data/config.inc.php.bak',
                       '/config/config_global.php.bak',
                       '/include/config.inc.php.tmp']

            for _url in url_lst:
                status, headers, html_doc = await self.http_request(_url)
                if status == 200 or status == 206:
                    if html_doc.find('<?php') >= 0:
                        await save_script_result(self, status, self.base_url + _url, 'Discuz Backup File Found')

            # getcolor DOM XSS
            status, headers, html_doc = await self.http_request('/static/image/admincp/getcolor.htm')
            if html_doc.find("if(fun) eval('parent.'+fun+'") > 0:
                await save_script_result(self, status, self.base_url + '/static/image/admincp/getcolor.htm',
                                   '', 'Discuz getcolor DOM XSS')



from lib.common import save_script_result

folders = """
/admin
/bak
/backup
/conf
/config
/db
/debug
/data
/database
/deploy
/WEB-INF
/install
/manage
/manager
/monitor
/tmp
/temp
/test
"""


async def do_check(self, url):
    if url != '/' or not self.conn_pool or self._404_status == 301:
        return

    _folders = folders.split()

    for _url in _folders:
        if not _url:
            continue
        status, headers, html_doc = await self.http_request(_url)

        if status in (301, 302):
            location = headers.get('location', '')
            if location.startswith(self.base_url + _url + '/') or location.startswith(_url + '/'):
                # save_user_script_result(self, status, self.base_url + _url,
                #                         '', 'Possible Sensitive Folder Found')
                await self.enqueue(_url + '/')
                await self.crawl(_url + '/')

        if status == 206 and self._404_status != 206 and headers.get('content-type', '').find('application/') >= 0:
            await save_script_result(self, status, self.base_url + _url, '', 'Possible Sensitive File Found')



from lib.common import save_script_result


async def do_check(self, url):
    if url == '/':
        if self.conn_pool and self.index_status in (301, 302):
            for keyword in ['admin', 'login', 'manage', 'backend']:
                if self.index_headers.get('location', '').find(keyword) >= 0:
                    await save_script_result(self, self.index_status, self.base_url + '/', 'Admin Site')
                    break



# -*- encoding: utf-8 -*-
# Generate HTML report

import asyncio
import time
from string import Template
import webbrowser
import sys
import codecs
import os
from lib.common import escape
from lib.consle_width import getTerminalSize
import lib.config as conf


html_head = """
<html>
<head>
<title>BBScan %s Scan Report</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<style>
    body {width: 90%%;min-width:960px; margin:auto; margin-top:10px; background:rgb(240,240,240);}
    p {color: #666;}
    h2 {color:#002E8C; font-size: 1em; padding-top:5px;}
    h1 {color:#0969da; font-size: 20px; padding-top:15px;}
    ul li {
    word-wrap: break-word;
    white-space: -moz-pre-wrap;
    white-space: pre-wrap;
    margin-bottom:10px;
    }
    span {color: aliceblue;}
    table thead tr th {
        text-align: left;
    }
        
    .leftAlign th,
    .leftAlign td {
        text-align: left !important;  
    }
    
    .purpleRed table {
        background-color: #3d3b63;
        border-collapse: collapse;
        border-radius: 10px;
        color: white;
        overflow: hidden;
        width: auto;
        /*max-width: 100%%;*/
    }
    .purpleRed thead {
        background-color: #f3646c;
        color: white;
    }
    .purpleRed th {
        letter-spacing: 0.3px;
        text-shadow: 0 1px 1px #2b2a466b;
    }
    .purpleRed table > thead > tr > th {
        border: 1px solid rgba(0, 0, 0, 0.1);
    }
    .purpleRed td {
        background-color: #3d3b63;
        color: white;
        font-size: 12px;
        word-break: break-all;
    }
    .purpleRed .vulRow td {
        background-color: #775bc9;
        color: white;
        font-size: 14px;
    }    
    .purpleRed td a {
      color: lightpink;
    }
    .purpleRed td strong {
        color: white;
    }
    .purpleRed td,
    .purpleRed th {
        border: 1px solid rgba(0, 0, 0, 0.1);
        padding: 11px;
    }
    .purpleRed thead th:hover,
    .purpleRed thead tr:hover,
    .purpleRed th:hover {
        background-color: #f3646c;
    }
    .purpleRed tbody tr:hover td,
    .purpleRed tbody tr:hover td a {
        background-color: #a32959;
        color: white;
    }

    a:link { 
      text-decoration: none; 
    } 
    a:visited { 
      text-decoration: none; 
    } 
    a:hover { 
      text-decoration: none; 
    } 
    a:active { 
      text-decoration: none; 
    }
    pre {
      overflow: auto;
      color: #ccffff;
      width:0;
      min-width:100%%;
    }
    pre::-webkit-scrollbar{
      height:6px;
      background-color: #ccffff;
      }
    pre::-webkit-scrollbar-thumb{
      background: #f3646c;
      border-radius:1px;
    }
    .apiRow td {
        background-color: darkslategrey;
    }
    .apiLinks {
        display: none;
        flex-wrap: wrap;
    }
    .apiLinks a {
        margin-right: 20px;
        margin-bottom: 5px;
    }
    button {padding-top: 3px; padding-bottom: 3px;}
</style>
<script>
    function switchShowHeaders(){
      btn = document.getElementById('switchBtn');
      if (btn.innerText === 'Hide Headers') {
          btn.innerText = 'Show Headers';
          display = 'none';
      } else {
          btn.innerText = 'Hide Headers';
          display = 'block';
      }
      for (let item of document.getElementsByTagName('pre')) {
        item.style.display = display;
      }
    }

    function showVulnerableTargets(){
      btn = document.getElementById('switchTargetBtn');
      if (btn.innerText === 'Show Vulnerable Targets Only') {
          btn.innerText = 'Show All Targets';
          all_rows = document.getElementsByTagName('tr');
          for (let i = 1; i < all_rows.length; i++) {
              if (all_rows[i].className.indexOf('vulRow') < 0 && all_rows[i].className.indexOf('apiRow') < 0) {
                  if (i === all_rows.length-1 || all_rows[i+1].className.indexOf('vulRow') < 0){
                      all_rows[i].style.display = 'none';
                  }
                  // api 行需要隐藏
                  if (i+1 <= all_rows.length-1 && all_rows[i+1].className.indexOf('apiRow') >= 0){
                        all_rows[i+1].style.display = 'none';
                  }
              }
          }
      } else {
          btn.innerText = 'Show Vulnerable Targets Only';
          for (let item of document.getElementsByTagName('tr')) {
              item.style.display = 'table-row';
          }
      }
    }
    function doSearch(reset){
        all_rows = document.getElementsByTagName('tr');
        keyword = document.getElementById('keyword').value;
        for (let i = 1; i < all_rows.length; i++) {
            item = all_rows[i];
            if (reset || item.innerText.indexOf(keyword) !== -1){
                item.style.display = 'table-row';
              } else {
                item.style.display = 'none';
              }
        }
    }
    function showLinks(linkID){
        element = document.getElementById(linkID);
        if (!element.style.display || element.style.display === 'none') {
            element.style.display = 'flex';
        } else {
            element.style.display = 'none';
        }
    }    
</script>
</head>
<body>

<h1>BBScan %s Scan Report</h1>
<p>Scanned <font color=red>${tasks_processed_count}</font> targets in 
<font color=green>${cost_min} ${cost_seconds} seconds</font>. 
<font color=red>${vulnerable_hosts_count}</font> vulnerable hosts found in total. </p>
<div style="padding-bottom: 10px;display: flex">
    <button id="switchBtn" onclick="switchShowHeaders()">Hide Headers</button>
    <button id="switchTargetBtn" onclick="showVulnerableTargets()" style="margin-left: 20px;margin-right: 30px;">Show Vulnerable Targets Only</button>
    <input id="keyword" type="text" style="width: 150px;outline:none;">
    <button id="searchBtn" onclick="doSearch()" style="margin-left: 10px;margin-right:10px;">Search</button>
    <button id="resetBtn" onclick="doSearch(true)">Reset</button>
    <button style="margin-left: auto; margin-right: 10px"><a href="https://github.com/lijiejie/BBScan/issues" target="_blank">Report Bugs</a></button>
</div>
<table class="purpleRed" style="width:100%%">
  <thead>
    <tr><th>Target</th><th>Fingerprint</th><th>Status</th><th>Web Server</th><th>Title</th><th width="40%%">Response Headers</th></tr>
  </thead>
""" % (conf.version, conf.version)

first_row = Template("""
     <tr> <td><a href="${url}" target="_blank">${url}</a></td> <td>${fingerprint}</td> <td>${status}</td> 
       <td>${server}</td> 
       <td><span>${title}</span></td> <td width="40%%"><pre>${headers}</pre></td> 
     </tr>
""")

vul_row = Template("""
     <tr class="vulRow"> <td colspan="2" style="text-align:right">${vul_type}</td> 
       <td>${status}</td> <td colspan="2" style="text-align:center"><span>${title}</span></td>
       <td><a href="${url}" target="_blank">${url}</a></td> 
     </tr>
""")

api_row = Template("""
     <tr class="apiRow"> 
     <td colspan="6"><button onclick="showLinks('${link_id}')">Display API Endpoints [ Found ${num_found} ]</button><div class="apiLinks" id="${link_id}">${api_urls}</div></td> 
     </tr>
""")

html_tail = """
</table>
<script>
    var input = document.getElementById('keyword');
    input.addEventListener("keypress", function(event) {
      if (event.key === "Enter") {
        event.preventDefault();
        document.getElementById("searchBtn").click();
      }
    });
</script>
</body>
</html>
"""


# template for markdown
markdown_general = """
# BBScan Scan Report
Version: %s
Num of targets: ${tasks_processed_count}
Num of vulnerable hosts: ${vulnerable_hosts_count}
Time elapsed: ${cost_min} ${cost_seconds} seconds
${content}
""" % conf.version

markdown_host = """
## ${host}
${list}
"""

markdown_list_item = """* [${status}] ${title} ${url}
"""

markdown = {
    'general': markdown_general,
    'host': markdown_host,
    'list_item': markdown_list_item,
    'suffix': '.md'
}


async def save_report(args, q_results, _file):

    start_time = time.time()

    report_dir = 'report/%s' % time.strftime('%Y%m%d', time.localtime())
    for dir_name in ['report', report_dir]:
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

    report_name = '%s_%s.html' % (os.path.basename(_file).lower().replace('.txt', ''),
                                  time.strftime('%H%M%S', time.localtime()))

    vulnerable_hosts_count = 0
    console_width = getTerminalSize()[0] - 2
    debug_file = codecs.open('debug.log', 'w', encoding='utf-8') if args.debug else None

    outfile = None
    api_links_id = 0

    try:
        while not conf.stop_me or q_results.qsize() > 0:
            if q_results.qsize() == 0:
                await asyncio.sleep(0.1)
                continue

            while q_results.qsize() > 0:
                item = await q_results.get()
                if type(item) is str:    # print msg only
                    message = '[%s] %s' % (time.strftime('%H:%M:%S', time.localtime()), item)
                    if args.debug:
                        debug_file.write(message + '\n')
                    # 对于大范围扫描情形，不换行引起滚屏，只在行内显示
                    if not args.debug and args.network <= 22 and \
                            (item.startswith('Scan ') or item.startswith('No ports open')):
                        sys.stdout.write(message + (console_width - len(message)) * ' ' + '\r')
                    else:
                        print(message)
                    continue

                host, results, urls_regex_found = item

                if not outfile:
                    outfile = codecs.open('%s/%s' % (report_dir, report_name), 'w', encoding='utf-8')
                    outfile.write(html_head)
                    conf.output_file_name = '%s/%s' % (report_dir, report_name)

                fingerprint = results.get('$Fingerprint', [''])[0]
                index = results.get('$Index', [{}])[0]
                index['fingerprint'] = '<br/>'.join(fingerprint)

                content = first_row.substitute(index)
                outfile.write(content)

                vuln_found = False
                for key in results.keys():
                    if key.startswith('$'):
                        continue
                    vuln_found = True
                    for item in results[key]:
                        item['vul_type'] = escape(item['vul_type'].replace('_', ' ')) if 'vul_type' in item else ''
                        content = vul_row.substitute(item)
                        outfile.write(content)

                if vuln_found:
                    vulnerable_hosts_count += 1

                if args.api and urls_regex_found:
                    api_links_id += 1
                    links = ''
                    for url in sorted(urls_regex_found):
                        if url.find('://') >= 0:
                            link = '<a href="%s" target="_blank">%s</a>' % (url, url)
                        elif url.lower().startswith('/'):
                            link = '<a href="%s" target="_blank">%s</a>' % (host + url, url)
                        else:
                            link = '<a href="%s" target="_blank">%s</a>' % (host + '/' + url, url)
                        links += link

                    content = api_row.substitute({'api_urls': links, 'num_found': len(urls_regex_found),
                                                  'link_id': 'apiLinks' + str(api_links_id)})
                    outfile.write(content)

        if args.debug and debug_file:
            debug_file.close()

        if conf.ports_saved_to_file:
            print('* Ports data saved to %s' % args.save_ports)

        if outfile:
            outfile.write(html_tail)
            outfile.close()
            with codecs.open('%s/%s' % (report_dir, report_name), 'r', encoding='utf-8') as f:
                content = f.read()
            cost_time = time.time() - start_time
            cost_min = int(cost_time / 60)
            cost_min = '%s min' % cost_min if cost_min > 0 else ''
            cost_seconds = '%.1f' % (cost_time % 60)
            content = content.replace('${tasks_processed_count}', str(conf.tasks_count)).replace(
                '${vulnerable_hosts_count}', str(vulnerable_hosts_count)).replace(
                '${cost_min}', str(cost_min)).replace('${cost_seconds}', str(cost_seconds))
            os.remove('%s/%s' % (report_dir, report_name))
            with codecs.open('%s/%s' % (report_dir, report_name), 'w', encoding='utf-8') as f:
                f.write(content)

            print('\n* %s vulnerable targets on sites in total.' % vulnerable_hosts_count)
            print('* Scan report saved to report/%s/%s' % (report_dir, report_name))
            if not args.no_browser:
                webbrowser.open_new_tab(os.path.abspath('%s/%s' % (report_dir, report_name)))
        else:
            print('\n* No vulnerabilities found on sites in %s.' % _file)

    except Exception as e:
        print('[save_report_thread Exception] %s %s' % (type(e), str(e)))
        import traceback
        traceback.print_exc()
        sys.exit(-1)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
#  Parse command line arguments
#


import argparse
import sys
import os
import glob
import re
import codecs
from lib.config import version


def parse_args():
    parser = argparse.ArgumentParser(prog='BBScan',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     description='* A fast vulnerability Scanner. *\n'
                                                 '* Find sensitive info disclosure vulnerabilities '
                                                 'from large number of targets *\n'
                                                 'By LiJieJie (https://www.lijiejie.com)',
                                     usage='BBScan.py [options]')

    group_target = parser.add_argument_group('Targets', '')
    group_target.add_argument('--host', metavar='HOST', type=str, default='', nargs='*',
                              help='Scan several hosts from command line')
    group_target.add_argument('-f', metavar='TargetFile', type=str, default='',
                              help='Load new line delimited targets from TargetFile')
    group_target.add_argument('-d', metavar='TargetDirectory', type=str, default='',
                              help='Load all *.txt files from TargetDirectory')
    group_target.add_argument('--crawler', metavar='CrawlDirectory', type=str, default='',
                              help='Load all *.log crawl files from CrawlDirectory')
    group_target.add_argument('--network', metavar='MASK', type=int, default=32,
                              help='Scan all Target/MASK neighbour hosts, \nshould be an integer between 8 and 31')
    group_target.add_argument('--skip', '--skip-intranet', dest='skip_intranet',
                              default=False, action='store_true',
                              help='Do not scan private IPs, when you are not under the same network with the target')

    group_http = parser.add_argument_group('Rule Based SCAN', '')
    group_http.add_argument('--rule', metavar='RuleFileName', type=str, default='', nargs='*',
                            help='Import specified rule files only.')
    group_http.add_argument('-n', '--no-crawl', dest='no_crawl', default=False, action='store_true',
                            help='No crawling, sub folders will not be processed')
    group_http.add_argument('--no-check404', dest='no_check404', default=False, action='store_true',
                            help='No HTTP 404 existence check')
    group_http.add_argument('--full', dest='full_scan', default=False, action='store_true',
                            help='Process all sub directories')
    group_http.add_argument('--fp', '--fingerprint', dest='fingerprint_only', default=False, action='store_true',
                            help='Disable rule and script scan, only check fingerprint')

    group_scripts = parser.add_argument_group('Script Based SCAN', '')
    group_scripts.add_argument('--scripts-only', dest='scripts_only', default=False, action='store_true',
                               help='Scan with user scripts only')
    group_scripts.add_argument('--script', metavar='ScriptName', type=str, default='', nargs='*',
                               help='Execute specified scripts only')
    group_scripts.add_argument('--no-scripts', dest='no_scripts', default=False, action='store_true',
                               help='Disable all scripts')

    group_concurrent = parser.add_argument_group('CONCURRENT', '')
    group_concurrent.add_argument('-p', metavar='PROCESS', type=int, default=30,
                                  help='Num of processes running concurrently, 30 by default')
    group_concurrent.add_argument('-t', metavar='THREADS', type=int, default=3,
                                  help='Num of scan threads for each scan process, 3 by default')

    group_other = parser.add_argument_group('OTHER', '')

    group_other.add_argument('--proxy', metavar='Proxy', type=str, default=None,
                             help='Set HTTP proxy server')

    group_other.add_argument('--timeout', metavar='Timeout', type=int, default=10,
                             help='Max scan minutes for each target, 10 by default')

    # Disabled for now, will be added back later
    # group_other.add_argument('-md', default=False, action='store_true',
    #                          help='Save scan report as markdown format')

    group_other.add_argument('--api', default=False, action='store_true',
                             help='Gather and display all API interfaces extracted from .js file')

    group_other.add_argument('--save-ports', metavar='PortsDataFile', dest='save_ports', type=str, default='',
                             help='Save open ports to PortsDataFile')

    group_other.add_argument('--debug', default=False, action='store_true',
                             help='Show verbose debug info')

    group_other.add_argument('--no-browser', dest='no_browser', default=False, action='store_true',
                             help='Do not open web browser to view report')

    group_other.add_argument('-v', action='version',
                             version='%(prog)s ' + version + ' (https://github.com/lijiejie/BBScan)')

    if len(sys.argv) == 1:
        sys.argv.append('-h')

    args = parser.parse_args()
    check_args(args)
    if args.f:
        args.input_files = [args.f]
    elif args.d:
        args.input_files = glob.glob(args.d + '/*.txt')
    elif args.crawler:
        args.input_files = ['crawler']
    elif args.host:
        args.input_files = ['hosts']

    return args


def check_args(args):
    if not (args.f or args.d or args.host or args.crawler):
        msg = 'Args missing! One of following args needs to be specified  \n' \
              '    -f TargetFile \n' \
              '    -d TargetDirectory \n' \
              '    --crawler TargetDirectory \n' \
              '    --host www.host1.com www.host2.com 8.8.8.8'
        print(msg)
        exit(-1)

    if args.f and not os.path.isfile(args.f):
        print('[ERROR] TargetFile not found: %s' % args.f)
        exit(-1)

    if args.d and not os.path.isdir(args.d):
        print('[ERROR] TargetDirectory not found: %s' % args.d)
        exit(-1)

    args.network = int(args.network)
    if not (8 <= args.network <= 32):
        print('[ERROR] Network should be an integer between 24 and 31')
        exit(-1)

    args.rule_files = []
    if args.rule:
        for rule_name in args.rule:
            if not rule_name.endswith('.txt'):
                rule_name += '.txt'
            if not os.path.exists('rules/%s' % rule_name):
                print('[ERROR] Rule file not found: %s' % rule_name)
                exit(-1)
            args.rule_files.append('rules/%s' % rule_name)

    args.require_no_http = True     # all scripts do not need http conn pool
    args.require_index_doc = False  # scripts need index html doc
    args.require_ports = set()      # ports need by scripts
    pattern = re.compile(r'ports_to_check.*?=(.*)')

    if not args.no_scripts:
        if args.script:
            for script_name in args.script:
                if not script_name.lower().endswith('.py'):
                    script_name += '.py'
                    if not os.path.exists('scripts/%s' % script_name):
                        print('* Script file not found: %s' % script_name)
                        exit(-1)

        for _script in glob.glob('scripts/*.py'):
            script_name_origin = os.path.basename(_script)
            script_name = script_name_origin.replace('.py', '')
            if args.script and script_name not in args.script and script_name_origin not in args.script:
                continue
            if script_name.startswith('_'):
                continue
            with codecs.open(_script, encoding='utf-8') as f:
                content = f.read()
                if content.find('self.http_request') > 0:
                    args.require_no_http = False
                if content.find('self.index_') > 0:
                    args.require_no_http = False
                    args.require_index_doc = True

                m = pattern.search(content)
                if m:
                    m_str = m.group(1).strip()
                    if m_str.find('#') > 0:  # remove comments
                        m_str = m_str[:m_str.find('#')]
                    if m_str.find('[') < 0:
                        if int(m_str) not in args.require_ports:
                            args.require_ports.add(int(m_str))
                    else:
                        for port in eval(m_str):
                            if port not in args.require_ports:
                                args.require_ports.add(int(port))

    # save open ports to file
    if args.save_ports:
        args.ports_file = None

    if args.proxy and args.proxy.find('://') < 0:
        args.proxy = 'http://%s' % args.proxy

    # 只需要指纹识别时，不需要404检查，也不需要抓取子页
    if args.fingerprint_only:
        args.no_check404 = True
        # args.no_crawl = True



# coding=utf-8

import asyncio
from lib.common import save_script_result

ports_to_check = 2181  # 默认服务端口


async def do_check(self, url):
    if url != '/':
        return
    port = 2181
    if self.scheme == 'zookeeper' and self.port != 2181:  # 非标准端口
        port = self.port
    elif 2181 not in self.ports_open:
        return

    try:
        reader, writer = await asyncio.open_connection(self.host, port)
        writer.write(b'envi')
        await writer.drain()
        data = await reader.read(1024)
        writer.close()
        if b'Environment' in data:
            await save_script_result(self, '', 'zookeeper://%s:%s' % (self.host, port), '', 'Zookeeper Unauthorized Access')
    except Exception as e:
        pass
    finally:
        try:
            writer.close()
            await writer.wait_closed()  # application data after close notify (_ssl.c:2730)
        except Exception as e:
            pass



# -*- encoding: utf-8 -*-

from lib.common import save_script_result
import httpx

ports_to_check = 8001    # 默认服务端口


async def do_check(self, url):
    if url != '/':
        return

    if self.conn_pool and self.index_headers.get('Server', '').startswith('kong/'):
        await save_script_result(self, '200', self.base_url, 'Kong Admin Rest API')

    if self.port == 8001:   # 如果已经维护了 8001 端口的 HTTP连接池，上面的逻辑已经完成扫描
        return

    if 8001 not in self.ports_open:    # 如果8001端口不开放
        return

    # 如果输入的是一个非标准端口的HTTP服务，需要单独对8001端口进行检测
    async with httpx.AsyncClient() as client:
        r = await client.get('http://%s:8001/' % self.host, follow_redirects=False, timeout=20)
        headers = r.headers
        if headers.get('Server', '').startswith('kong/'):
            await save_script_result(self, r.status_code, 'http://%s:8001' % self.host, 'Kong Admin Rest API')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import asyncio
from lib.common import save_script_result


ports_to_check = 6379    # 默认扫描端口


async def do_check(self, url):
    if url != '/':
        return
    port = 6379
    # 非标准端口，不需要检查6379端口是否开放
    # 支持用户传入目标 redis://test.ip:16379 来扫描非标准端口上的Redis服务
    if self.scheme == 'redis' and self.port != 6379:
        port = self.port
    elif 6379 not in self.ports_open:
        return

    try:
        host = self.host.split(':')[0]
        reader, writer = await asyncio.open_connection(host, port)
        payload = b'\x2a\x31\x0d\x0a\x24\x34\x0d\x0a\x69\x6e\x66\x6f\x0d\x0a'
        writer.write(payload)
        await writer.drain()
        data = await reader.read(1024)
        writer.close()
        if b"redis_version" in data:
            await save_script_result(self, '', 'redis://%s:%s' % (host, port), 'Redis Unauthorized Access')
    except Exception as e:
        pass
    finally:
        try:
            writer.close()
            await writer.wait_closed()  # application data after close notify (_ssl.c:2730)
        except Exception as e:
            pass



# /{hostname_or_folder}.zip         {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.rar         {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.tar.gz      {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.tar.bz2     {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.tgz         {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.7z          {status=206}    {type="application/"}   {root_only}
# /{hostname_or_folder}.log         {status=206}    {type="application/"}   {root_only}
#
# /{sub}.zip         {status=206}    {type="application/"}   {root_only}
# /{sub}.rar         {status=206}    {type="application/"}   {root_only}
# /{sub}.tar.gz      {status=206}    {type="application/"}   {root_only}
# /{sub}.tar.bz2     {status=206}    {type="application/"}   {root_only}
# /{sub}.tgz         {status=206}    {type="application/"}   {root_only}
# /{sub}.7z          {status=206}    {type="application/"}   {root_only}
#
# /../{hostname_or_folder}.zip         {status=206}    {type="application/"}
# /../{hostname_or_folder}.rar         {status=206}    {type="application/"}
# /../{hostname_or_folder}.tar.gz      {status=206}    {type="application/"}
# /../{hostname_or_folder}.tar.bz2     {status=206}    {type="application/"}
# /../{hostname_or_folder}.tgz         {status=206}    {type="application/"}
# /../{hostname_or_folder}.7z          {status=206}    {type="application/"}
# /../{hostname_or_folder}.log         {status=206}    {type="application/"}


from lib.common import save_script_result


async def do_check(self, url):
    if not self.conn_pool:
        return

    extensions = ['.zip', '.rar', '.tar.gz', '.tar.bz2', '.tgz', '.7z', '.log', '.sql']

    if url == '/' and self.domain_sub:
        file_names = [self.host.split(':')[0], self.domain_sub]
        for name in file_names:
            for ext in extensions:
                status, headers, html_doc = await self.http_request('/' + name + ext)
                if status in [206, 200]:
                    if ext == '.sql' and html_doc.find("CREATE TABLE") >= 0:
                        await save_script_result(self, status, self.base_url + '/' + name + ext, '', 'Compressed File')
                    elif headers.get('content-type', '').find('application/') >= 0:
                        status2, headers2, html_doc2 = await self.http_request('/fptest' + name + ext)
                        if status2 in [206, 200] and headers2.get('content-type', '').find('application/') >= 0:
                            pass
                        else:
                            status3, headers3, html_doc3 = await self.http_request('/' + name + 'fptest' + ext)
                            if status3 in [206, 200] and headers3.get('content-type', '').find('application/') >= 0:
                                pass
                            else:
                                await save_script_result(self, status, self.base_url + '/' + name + ext, '',
                                                         'Compressed File')

    elif url != '/' and len(url.split('/')) >= 2:
        # sub folders like /aaa/bbb/
        folder_name = url.split('/')[-2]
        if len(folder_name) >= 4:
            url_prefix = url[: -len(folder_name) - 1]
            for ext in extensions:
                status, headers, html_doc = await self.http_request(url_prefix + folder_name + ext)
                if status == 206 and headers.get('content-type', '').find('application/') >= 0:
                    status2, headers2, html_doc2 = await self.http_request(url_prefix + folder_name + 'fptest' + ext)
                    if status2 == 206 and headers2.get('content-type', '').find('application/') >= 0:
                        pass
                    else:
                        await save_script_result(self, status, self.base_url + url_prefix + folder_name + ext,
                                                 '', 'Compressed File')
